﻿// Goldie: GOLD Engine for D
// Written in the D programming language.

module /+S:PACKAGE+/goldie/+E:PACKAGE+/.lang;

/+P:SET_VERSION+/

import tango.core.Array;
import tango.io.Stdout;
import tango.io.device.File;
import tango.text.convert.UnicodeBom;

import semitwist.util.all;
import semitwist.util.compat.all;

import goldie.base;
import goldie.file;
import goldie.lexer;
import goldie.parser;
import goldie.token;

version(GoldieStaticLang)
{
	import goldie.lang;
	
	import /+P:PACKAGE+/DUMMY.lexer;
	import /+P:PACKAGE+/DUMMY.parser;

	public /+P:LANG_CLASSNAME+/DUMMY /+P:LANG_INSTNAME+/DUMMY;
	static this()
	{
		/+P:LANG_INSTNAME+/DUMMY = new /+P:LANG_CLASSNAME+/DUMMY();
	}
}

/+S:LANG_INHERIT+/public class Language/+E:LANG_INHERIT+/
{
	/+S:REM+/
	mixin(getter!("protected", string, "name"));
	mixin(getter!("protected", string, "ver"));
	mixin(getter!("protected", string, "author"));
	mixin(getter!("protected", string, "about"));
	mixin(getter!("protected", bool,   "caseSensitive"));
	mixin(getter!("protected", string, "filename"));
	
	protected Symbol[][string] symbolLookup;
	
	mixin(getter!("package", Symbol[],    "symbolTable"));
	mixin(getter!("package", CharSet[],   "charSetTable"));
	mixin(getter!("package", Rule[],      "ruleTable"));
	mixin(getter!("package", DFAState[],  "dfaTable"));
	mixin(getter!("package", LALRState[], "lalrTable"));
	
	mixin(getter!("protected", int, "startSymbolIndex"));
	mixin(getter!("protected", int, "initialDFAState"));
	mixin(getter!("protected", int, "initialLALRState"));
	
	mixin(getter!("protected", int, "eofSymbolIndex"));
	mixin(getter!("protected", int, "errorSymbolIndex"));

	protected Lexer lexer;
	protected Parser parser;

	package this()
	{
	}
	
	package this(string cgtFilename)
	{
		auto goldFile = new CGTFile(cgtFilename);
		filename = cgtFilename;
		
		this.symbolLookup = goldFile.symbolLookup;
		mixin(initMemberX!("this._{} = goldFile.{}",
			name, ver, author, about, caseSensitive,
			startSymbolIndex, initialDFAState, initialLALRState,
			eofSymbolIndex, errorSymbolIndex,
			symbolTable, charSetTable, ruleTable,
			dfaTable, lalrTable
		));
	}
	/+E:REM+/
	
	/+P:LANG_STATICDATA+/
	
	version(GoldieStaticLang)
	{
		private this()
		{
			super();
			/+P:LANG_INIT+/
		}

		public /+P:LEXER_CLASSNAME+/DUMMY lexCode(string source, string filename="")
		{
			return cast(/+P:LEXER_CLASSNAME+/DUMMY)lexCodeX(source, filename);
		}
		
		public /+P:PARSER_CLASSNAME+/DUMMY parseTokens(Token[] tokens, string filename="", Lexer lexerUsed=null)
		{
			return cast(/+P:PARSER_CLASSNAME+/DUMMY)parseTokensX(tokens, filename, lexerUsed);
		}

		public /+P:LEXER_CLASSNAME+/DUMMY lexFile(string filename)
		{
			return cast(/+P:LEXER_CLASSNAME+/DUMMY)lexFileX(filename);
		}

		public /+P:PARSER_CLASSNAME+/DUMMY parseFile(string filename)
		{
			return cast(/+P:PARSER_CLASSNAME+/DUMMY)parseFileX(filename);
		}

		public /+P:PARSER_CLASSNAME+/DUMMY parseCode(string source, string filename="")
		{
			return cast(/+P:PARSER_CLASSNAME+/DUMMY)parseCodeX(source, filename);
		}
	}

	public Lexer lexCodeX(string source, string filename="")
	{
		lexer = new /+S:LEXER_CLASSNAME+/Lexer/+E:LEXER_CLASSNAME+/();
		lexer.process(source, this, filename);
		return lexer;
	}
	
	public Parser parseTokensX(Token[] tokens, string filename="", Lexer lexerUsed=null)
	{
		parser = new /+S:PARSER_CLASSNAME+/Parser/+E:PARSER_CLASSNAME+/();
		parser.process(tokens, this, filename, lexerUsed);
		return parser;
	}

	/+S:REM+/
	public Lexer lexFileX(string filename)
	{
		auto bom = new UnicodeBom!(char)(Encoding.Unknown);
		return lexCodeX(bom.decode(File.get(filename)), filename);
	}

	public Parser parseFileX(string filename)
	{
		return parseTokensX(lexFileX(filename).tokens, filename, lexer);
	}

	public Parser parseCodeX(string source, string filename="")
	{
		return parseTokensX(lexCodeX(source, filename).tokens, filename, lexer);
	}

	protected Symbol[] lookupSymbol(string symName)
	{
		if(symName in symbolLookup)
			return symbolLookup[symName];
		else
			return [];
	}

	public Symbol[] symbolsByName(string name)
	{
		return lookupSymbol(name).dup;
	}
	
	public SymbolType[] symbolTypesByName(string name)
	{
		return
			lookupSymbol(name).map(
				(Symbol sym) { return sym.type; }
			);
	}
	
	public string symbolTypesStrByName(string name)
	{
		return symbolTypesByName(name).symbolTypesToString();
	}
	
	//TODO: This is just a quick-n-dirty implementation atm, probably runs slow
	public int ruleIdOf(string parentSymbol, string[] subSymbols...)
	{
		void throwNoSuchRule()
		{
			throw new Exception("Rule does not exist: {} ::= {}".sformat(parentSymbol, subSymbols));
		}
		
		foreach(string symName; subSymbols)
		{
			if(isSymbolNameAmbiguous(symName))
				throw new Exception(
					"Symbol '{}' is ambiguous, it could be any of the following types: {}\nGoldie does not yet support disambiguation of symbol names at runtime."
						.sformat(symName, symbolTypesStrByName(symName))
				);
				
			if(!isSymbolNameValid(symName))
				throwNoSuchRule();
		}
		
		foreach(int ruleId, Rule rule; ruleTable)
		{
			// Wrong parentSymbol?
			if(rule.symbolIndex != lookupSymbol(parentSymbol)[0].id)
				continue;
			
			// Wrong subSymbols length?
			if(subSymbols.length != rule.subSymbolIndicies.length)
				continue;

			if(rule.subSymbolIndicies.length == 0)
			{
				// Require an explicit null for consistency with static-mode
				if(subSymbols.length == 1 && subSymbols[0] is null)
					return ruleId;
				else
					continue;
			}

			bool foundMatch = true;
			foreach(int i, int subSymbolIndex; rule.subSymbolIndicies)
			if(subSymbolIndex != lookupSymbol(subSymbols[i])[0].id)
			{
				foundMatch = false;
				break;
			}
			
			if(foundMatch)
				return ruleId;
		}

		throwNoSuchRule();
		assert(0);
	}
	/+E:REM+/
	
	public /+P:STATIC+/ bool /+S:TO:STATIC_IDENT+/isSymbolNameValid/+E:TO:STATIC_IDENT+/(string name)
	{
		return /+S:TO:STATIC_IDENT+/lookupSymbol/+E:TO:STATIC_IDENT+/(name).length >= 1;
	}
	
	public /+P:STATIC+/ bool /+S:TO:STATIC_IDENT+/isSymbolNameAmbiguous/+E:TO:STATIC_IDENT+/(string name)
	{
		return /+S:TO:STATIC_IDENT+/lookupSymbol/+E:TO:STATIC_IDENT+/(name).length > 1;
	}
	
	public /+P:STATIC+/ string[] /+S:TO:STATIC_IDENT+/uniqueSymbolNames/+E:TO:STATIC_IDENT+/()
	{
		version(GoldieStaticLang)
			return staticUniqueSymbolNameArray;
		else
			return symbolLookup.keys;
	}
	
	public /+P:STATIC+/ Symbol /+S:TO:STATIC_IDENT+/eofSymbol/+E:TO:STATIC_IDENT+/()
	{
		version(GoldieStaticLang)
			return staticSymbolTable[staticEofSymbolIndex];
		else
			return symbolTable[eofSymbolIndex];
	}
	
	public /+P:STATIC+/ Symbol /+S:TO:STATIC_IDENT+/errorSymbol/+E:TO:STATIC_IDENT+/()
	{
		version(GoldieStaticLang)
			return staticSymbolTable[staticErrorSymbolIndex];
		else
			return symbolTable[errorSymbolIndex];
	}
	
	version(GoldieStaticLang)
	{
		public bool isSymbolNameValid(string name)
		{
			return typeof(this)./+S:TO:STATIC_IDENT+/isSymbolNameValid/+E:TO:STATIC_IDENT+/(name);
		}
		
		public bool isSymbolNameAmbiguous(string name)
		{
			return typeof(this)./+S:TO:STATIC_IDENT+/isSymbolNameAmbiguous/+E:TO:STATIC_IDENT+/(name);
		}
		
		public string[] uniqueSymbolNames()
		{
			return typeof(this)./+S:TO:STATIC_IDENT+/uniqueSymbolNames/+E:TO:STATIC_IDENT+/();
		}
		
		public Symbol eofSymbol()
		{
			return typeof(this)./+S:TO:STATIC_IDENT+/eofSymbol/+E:TO:STATIC_IDENT+/();
		}
		
		public Symbol errorSymbol()
		{
			return typeof(this)./+S:TO:STATIC_IDENT+/errorSymbol/+E:TO:STATIC_IDENT+/();
		}

		protected static Symbol[] /+S:TO:STATIC_IDENT+/lookupSymbol/+E:TO:STATIC_IDENT+/(string symName)
		{
			switch(symName)
			{
			/+P:LOOKUP_SYMBOL+/
			}
		}

		protected Symbol[] lookupSymbol(string symName)
		{
			return typeof(this)./+S:TO:STATIC_IDENT+/lookupSymbol/+E:TO:STATIC_IDENT+/(symName);
		}
	}
}
