﻿// Goldie: GOLD Engine for D
// Samples: Dynamic-Language Calculator
// Written in the D programming language.

/++
A simple arithmetic calculator utilizing a calculator grammar
compiled by GOLD.

Uses a dynamically loaded language and dynamic-style tokens.

$(WEB http://www.dsource.org/projects/goldie, Goldie Homepage)

Author:
$(WEB www.semitwist.com, Nick Sabalausky)

This has been tested to work with:
  - DMD 1.056 / Tango 0.99.9 / Rebuild 0.76
  - DMD 1.056 / Tango 0.99.9 / xfBuild 0.4
+/

module samples.calculatorDynamic.main;

import tango.io.Stdout;
import tango.text.Unicode;
import tango.util.Convert;

import semitwist.util.all : sformat, getExecPath;
import semitwist.cmd.plain;
import semitwist.util.compat.all;

import goldie.all;

real calculate(Token tok)
{
	if(tok.matches("Number"))
		return to!(real)(tok.toString());
		

	else if(tok.matches("<Add Exp>", "<Mult Exp>"))
		return calculate(tok.subX[0]);

	else if(tok.matches("<Add Exp>", "<Add Exp>", "+", "<Mult Exp>"))
		return calculate(tok.subX[0]) + calculate(tok.subX[2]);

	else if(tok.matches("<Add Exp>", "<Add Exp>", "-", "<Mult Exp>"))
		return calculate(tok.subX[0]) - calculate(tok.subX[2]);


	else if(tok.matches("<Mult Exp>", "<Negate Exp>"))
		return calculate(tok.subX[0]);
			
	else if(tok.matches("<Mult Exp>", "<Mult Exp>", "*", "<Negate Exp>"))
		return calculate(tok.subX[0]) * calculate(tok.subX[2]);
		
	else if(tok.matches("<Mult Exp>", "<Mult Exp>", "/", "<Negate Exp>"))
		return calculate(tok.subX[0]) / calculate(tok.subX[2]);
		
		
	else if(tok.matches("<Negate Exp>", "<Value>"))
		return calculate(tok.subX[0]);

	else if(tok.matches("<Negate Exp>", "-", "<Value>"))
		return -calculate(tok.subX[1]);
	
	
	else if(tok.matches("<Value>", "Number"))
		return calculate(tok.subX[0]);

	else if(tok.matches("<Value>", "(", "<Add Exp>", ")"))
		return calculate(tok.subX[1]);

		
	else
		throw new Exception("Unhandled token: {}".sformat(tok.name));
}

int main(string[] args)
{
	// This is the directory where Goldie will look for cgt files,
	// (But it's not needed if you use Goldie.loadCGT, which takes
	// an ordinary file/path just like any normal file function).
	//
	// The default for this is "{the exec file}/lang/".
	// Note: Forward slashes will work here on both Windows and Unix.
	Goldie.langDir = getExecPath()~"../lang/";

	Stdout("Calculator").newline;
	
	Stdout("Loading language...").flush;
	Language lang = Goldie.loadLanguage("calc");
	Stdout("done!").newline;

	Stdout("Enter an arithmetic expression ('exit' to exit)").newline;
	while(true)
	{
		Stdout.newline;
		string src = cmd.prompt("calculator>");
		
		if(src.toLower() == "exit")
			break;
			
		try
		{
			auto parseTree = lang.parseCodeX(src).parseTreeX;
			real result = calculate(parseTree);
			Stdout.formatln("{:d5}", result);
		}
		catch(Exception e)
			Stdout.formatln("{}", e);
	}
	
	return 0;
}
